from typing import Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.combination import Reference


class DomainStep(BaseStep):
    name = Field('domain', const=True)
    domain: Union[str, Reference]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import TextStep


def translate_text(
    step: TextStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'TEXT_STEP_{index}'
    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step.text: {step.text}\n'
        f'step.new_column: {step.new_column}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str()
    renamed_fields = f"""'{step.text}' AS {step.new_column} """
    if completed_fields:
        renamed_fields = f', {renamed_fields}'
    transformed_query = (
        f"""{query.transformed_query}, {query_name} AS"""
        f""" (SELECT {completed_fields}{renamed_fields}"""
        f"""FROM {query.query_name})"""
    )

    query.metadata_manager.add_query_metadata_column(step.new_column, 'text')

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from weaverbird.pipeline.steps import TextStep


def translate_text(step: TextStep) -> list:
    return [
        {
            '$addFields': {step.new_column: {'$literal': step.text}},
        }
    ]



from typing import Any, Callable, List, Literal, Optional, Protocol, Tuple

from pandas import DataFrame
from pydantic import BaseModel

from weaverbird.backends.sql_translator.metadata import SqlQueryMetadataManager
from weaverbird.pipeline import Pipeline


class SQLStepTranslationReport(BaseModel):
    step_index: int


class SQLPipelineTranslationReport(BaseModel):
    sql_steps_translation_reports: List[SQLStepTranslationReport]


class TableMetadataUpdateError(Exception):
    ...


class SQLQuery(BaseModel):
    query_name: Optional[str]
    transformed_query: Optional[str]
    selection_query: Optional[str]
    metadata_manager: Optional[SqlQueryMetadataManager]


SQLQueryRetriever = Callable[[str], str]
SQLQueryDescriber = Callable
SQLQueryExecutor = Callable[[str, str], DataFrame]

SQLPipelineTranslator = Callable[
    [Pipeline, SQLQueryRetriever, SQLQueryDescriber, SQLQueryExecutor],
    Tuple[str, SQLPipelineTranslationReport],
]
SQLDialect = Literal['snowflake', 'postgres']


class SQLStepTranslator(Protocol):
    def __call__(
        self,
        step: Any,
        query: SQLQuery,
        index,
        sql_query_retriever: Optional[SQLQueryRetriever],
        sql_query_describer: Optional[SQLQueryDescriber],
        sql_query_executor: Optional[SQLQueryExecutor],
        sql_translate_pipeline: Optional[SQLPipelineTranslator],
        subcall_from_other_pipeline_count: Optional[int],
        sql_dialect: Optional[SQLDialect],
    ):
        ...



from abc import ABC
from datetime import datetime
from typing import Any, List, Literal, Union

from pydantic import BaseModel, Field

from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames


class BaseCondition(BaseModel):
    ...


class ComparisonCondition(BaseCondition):
    column: ColumnName
    operator: Literal['eq', 'ne', 'lt', 'le', 'gt', 'ge']
    value: Any


class InclusionCondition(BaseCondition):
    column: ColumnName
    operator: Literal['in', 'nin']
    value: List[Any]


class NullCondition(BaseCondition):
    column: ColumnName
    operator: Literal['isnull', 'notnull']


class MatchCondition(BaseCondition):
    column: ColumnName
    operator: Literal['matches', 'notmatches']
    value: str


class DateBoundCondition(BaseModel):
    column: ColumnName
    operator: Literal['from', 'until']
    value: Union[RelativeDate, datetime, str]


SimpleCondition = Union[
    ComparisonCondition,
    InclusionCondition,
    NullCondition,
    MatchCondition,
    DateBoundCondition,
]


class BaseConditionCombo(BaseCondition, ABC):
    class Config(PopulatedWithFieldnames):
        ...

    def to_dict(self):
        return self.dict(by_alias=True)


class ConditionComboAnd(BaseConditionCombo):
    and_: List['Condition'] = Field(..., alias='and')


class ConditionComboOr(BaseConditionCombo):
    or_: List['Condition'] = Field(..., alias='or')


Condition = Union[ConditionComboAnd, ConditionComboOr, SimpleCondition]
ConditionComboOr.update_forward_refs()
ConditionComboAnd.update_forward_refs()



from typing import Any, Callable, Dict

from weaverbird.backends.mongo_translator.steps.argmax import translate_argmax
from weaverbird.backends.mongo_translator.steps.argmin import translate_argmin
from weaverbird.backends.mongo_translator.steps.domain import translate_domain
from weaverbird.backends.mongo_translator.steps.filter import translate_filter
from weaverbird.backends.mongo_translator.steps.text import translate_text
from weaverbird.pipeline.steps import BaseStep

# I would like to have a better type for the callable, but I don't know how to do it.
# each of this function take a particular step as input
# so it is not possible to use `Dict[str, Callable[[BaseStep], list]]
mongo_step_translator: Dict[str, Callable[[Any], list]] = {
    'argmax': translate_argmax,
    'argmin': translate_argmin,
    'domain': translate_domain,
    'text': translate_text,
    'filter': translate_filter,
}
